program main
  !solve the 1D nonlinear Poisson-Vlasov system via the delta-f PIC method. The document for this program can be found at http://theory.ipp.ac.cn/~yj/research_notes/particle_simulation.pdf
  !these codes derive from the codes in ~/project_new/pic_full-f, which is a 1D full-f PIC code
  !external library used: FFTW
  !written by Youjun Hu (yjhu@ipp.cas.cn) of Institute of Plasma Physics, Chinese Academy of Sciences
  use  precision,only:p_
  use  constants,only:zero,one,two,four,five,twopi,eight
  use spatial_grids,only:ngrids,width,x,dx,xmin,xmax,rho,phi,efield !phi is the electric potential, efield is the electric field
  use omp_lib !openmp libriary
  implicit none
  integer,parameter:: loading_scheme=2 !the results given by loading_scheme=1 and loading_scheme=2 agree with each other.
  integer,parameter::total_number=200000 !number of sampling points in phase space
  real(p_):: r(total_number),v(total_number) !position and velocity of partilces
  real(p_):: particle_weight(total_number),particle_weight0(total_number),ps_vol(total_number) !ps_vol(j) is the phase space volume occupied by the jth particle marker.
  real(p_):: deltaf(total_number),deltaf0(total_number)
  real(p_):: r_old(total_number),v_old(total_number) !position and velocity of partilces 
  real(p_):: r_mid(total_number),v_mid(total_number) !position and velocity of partilces 
  real(p_):: deltaf_old(total_number)
  real(p_):: full_f(total_number)
  real(p_):: f_zero !function name
  integer,parameter:: ngrids_val=400 !please choose a even number because FFT algorithm assume this
  real(p_),parameter:: width_val=100._p_ !in unit of Debye length
  real(p_),parameter:: dt=0.1_p_  !in unit of 1/wpe, where wpe is the electron plasma frequency
  integer,parameter:: nstep=101*2,kplot=100 !every ksplot steps, store data for plotting
  real(p_):: t
  character(100):: file1,file2,file3,file4
  real(p_):: tarray(2) !store the cpu clock time
  integer:: i,k

  INTEGER :: count1,count2, count_rate, count_max
  CALL SYSTEM_CLOCK(count1, count_rate, count_max)

  call cpu_time(tarray(1))    !cpu_time is a f95 intrinsic subroutine

  ngrids=ngrids_val
  width=width_val
  allocate(x(0:ngrids))
  allocate(phi(0:ngrids))
  allocate(efield(0:ngrids))
  allocate(rho(0:ngrids))

  dx=width/ngrids !this is correct
  do i=0,ngrids !create_spatial_grids
     x(i)=0._p_+dx*i
  enddo
  xmin=x(0)
  xmax=x(ngrids)

  call loading_particles(loading_scheme,total_number,r,v,ps_vol) !initial sampling of the phase space

  !deltaf=0._p_ 
  call set_deltaf0(total_number,r,v,deltaf0)
  deltaf=deltaf0

  particle_weight=deltaf*ps_vol
  particle_weight0= particle_weight

  do i=1,total_number
     full_f(i)=f_zero(r(i),v(i))+deltaf0(i)
  enddo


  t=0._p_
  file1='r000.txt'
  file2='v000.txt'
  file3='r_v000.txt'
  file4='sf000.txt'
  open(101,file='energy_evolution.txt')
  open(102,file='fourier_evolution.txt')

  do k=1,nstep
     r_old=r
     v_old=v
     deltaf_old=deltaf

     call density(total_number,particle_weight,r,x,rho,ngrids) !use PIC method to calculate the charge density at grids
     !call poisson_solver_fd(ngrids,dx,rho,phi) !solve Poisson's equation to get the electric potential phi

!!$     call poisson_solver_dft(ngrids,dx,rho,phi) !solve Poisson's equation using FFT method 
!!$     phi(ngrids)=phi(0) !periodic boundary condition
!!$     call calculate_efield(ngrids,phi,dx,x,efield)
     call efield_solver_dft(ngrids,dx,rho,efield) !solve Poisson's equation using FFT method 
     efield(ngrids)=efield(0) !periodic boundary condition

     !$omp parallel do
     do i=1,total_number
       call half_push(r(i),v(i),deltaf(i),dt) !push particles and deltaf for half time-step (first step of the 2nd Runge-Kutta method, Euler scheme)
       !deltaf(i)=full_f(i)-f_zero(r(i),v(i)) !using the simple formual to update deltaf, insteal of using time integration of the time derivative of deltaf
     enddo
     !$omp end parallel do
     particle_weight=deltaf*ps_vol
     !--------------------------------------------------------

     !----second step of the 2nd Runge-Kutta method
     call density(total_number,particle_weight,r,x,rho,ngrids) !use PIC method to calculate the charge density at grids, at half-time-step
     call efield_solver_dft(ngrids,dx,rho,efield) !solve Poisson's equation at half-time-step
     efield(ngrids)=efield(0) !periodic boundary condition
     
     r_mid=r !value at half step
     v_mid=v !value at half step
     
     r=r_old !go back to the starting point
     v=v_old !go back to the starting point
     deltaf=deltaf_old !go back to the starting point

     !$omp parallel do
     do i=1,total_number
        call full_push(r_mid(i),v_mid(i),r(i),v(i),deltaf(i),dt) !second step of 2nd Runge-Kutta method, push particles and delta-f for one full time-step using the values at half-time-step
        !deltaf(i)=full_f(i)-f_zero(r(i),v(i)) !using the simple formual to update deltaf, insteal of using time integration of the time derivative of deltaf
     enddo
     !$omp end parallel do
     particle_weight=deltaf*ps_vol
     !----------------------------------------

     if (mod(k-1,kplot).eq. 0) then
        write(file1(2:4),'(i3.3)') (k-1)/kplot
        write(file2(2:4),'(i3.3)') (k-1)/kplot
        call reduced_distribution(total_number,r,v,particle_weight,file1,file2)
        write(file3(4:6),'(i3.3)') (k-1)/kplot
        call phase_space_structure(r,v,deltaf,total_number,file3)
        write(file4(3:5),'(i3.3)') (k-1)/kplot
        call record_field(ngrids,x,rho,phi,efield,file4)
     endif

     t=t+dt
     !write(*,*) (k-1)*dt,rho(ngrids/3),phi(ngrids/3),efield(ngrids/3)
     call energy_evolution(t,total_number,v,particle_weight,efield,ngrids)
     call fourier_evolution(t,efield,ngrids)

  enddo !------main time-advancing
  close(101)
  close(102)
  call identify_rp(total_number,r,v,deltaf0,deltaf) !identify resonant particles
  call cpu_time(tarray(2))
  write (*,*) 'CPU time used (seconds)', tarray(2)-tarray(1)
  CALL SYSTEM_CLOCK(count2, count_rate, count_max)
  write(*,*) 'Wall time used (seconds) ', (count2-count1)/count_rate
end program main




subroutine half_push(r,v,deltaf,dt)
  use precision,only:p_
  use constants,only:zero,one,one_half,three,six
  use spatial_grids,only:width,xmin,xmax
  implicit none

  real(p_),intent(in):: dt
  real(p_),intent(inout):: r,v,deltaf
  real(p_):: r_dot,v_dot,deltaf_dot !equations of motion

!Euler-scheme
  r=r+r_dot(r,v)*dt*one_half
  v=v+v_dot(r,v)*dt*one_half
  deltaf=deltaf+deltaf_dot(r,v)*dt*one_half

end subroutine half_push


subroutine full_push(r_mid,v_mid,r,v,deltaf,dt)
  use precision,only:p_
  use constants,only:zero,one,one_half,three,six
  use spatial_grids,only:width,xmin,xmax
  implicit none

  real(p_),intent(in):: dt,r_mid,v_mid
  real(p_),intent(inout):: r,v,deltaf
  real(p_):: r_dot,v_dot,deltaf_dot !equations of motion

  r=r+r_dot(r_mid,v_mid)*dt
  v=v+v_dot(r_mid,v_mid)*dt !electric field used is also evaluated at the half-time-step
  deltaf=deltaf+deltaf_dot(r_mid,v_mid)*dt !electric field used is also evaluated at the half-time-step

end subroutine full_push


function r_dot(r,v)
  use precision,only:p_
  implicit none
  real(p_):: r_dot,r,v
  r_dot=v
end function r_dot


function v_dot(r,v)
  use precision,only:p_
  use constants,only: one,zero
  use spatial_grids,only: x,ngrids,efield,dx,width,xmin,xmax
  implicit none
  real(p_):: v_dot,r,v
  integer:: nloc
  real(p_):: w,ef
  if (r>xmax)  r=r-width !shift into the computational region
  if (r<xmin)   r=r+width !shift into the computational region
!  call locate(r,x,ngrids,nloc) ! r is between nloc and nloc+1
call location(ngrids,x,r,nloc) ! r is between nloc and nloc+1
! if(nloc.eq.ngrids-1) write(*,*) 'r=',r
  w=(r-x(nloc))/dx
  ef=efield(nloc)*(one-w)+efield(nloc+1)*w
  v_dot=-ef !nonlinear case
!  v_dot=0._p_ !to test linear case
end function v_dot


function deltaf_dot(r,v)
  use precision,only:p_
  use constants,only: one,zero
  use spatial_grids,only: x,ngrids,efield,dx,width,xmin,xmax
  implicit none
  real(p_):: deltaf_dot,r,v
  real(p_):: df0dv !the gradient of f0 with respect to v
  integer:: nloc
  real(p_):: w,ef

  call location(ngrids,x,r,nloc) ! r is between nloc and nloc+1
  w=(r-x(nloc))/dx
  ef=efield(nloc)*(one-w)+efield(nloc+1)*w
  deltaf_dot=ef*df0dv(v)
end function deltaf_dot

function df0dv(v)
  use precision,only:p_
  use constants,only: one,zero,two,twopi,pi
  use velocity_space,only: vb
  implicit none
  real(p_):: v,df0dv
!  df0dv=exp(-v*v/two)*(-v)/sqrt(twopi) !for Maxwelian distribution,wrong!
  df0dv=exp(-v*v)*(-2.*v)/sqrt(pi) !for Maxwelian distribution
!  df0dv=0.5*(exp(-(v-vb)**2/two)*(-(v-vb)+exp(-(v+vb)**2/two)*(-(v+vb))))/sqrt(twopi) !for Maxwelian distribution,wrong!
end function df0dv


function f_zero(r,v)
  use precision,only:p_
  use constants,only: one,zero,two,twopi,pi
  use velocity_space,only: vb
  implicit none
  real(p_):: v,r,f_zero

  f_zero=exp(-v*v)/sqrt(pi) !for Maxwelian distribution

end function f_zero


subroutine set_deltaf0(total_number,r,v,deltaf0)
  use  precision,only:p_
  use constants,only: twopi,pi
  use spatial_grids,only: width
  implicit none
  integer,intent(in):: total_number
  real(p_),intent(in):: r(total_number),v(total_number)
  real(p_),intent(inout):: deltaf0(total_number)
  integer:: i
  real(p_):: random_yj
  real(p_):: kvect

!    kvect=2*twopi/width
   kvect=8*twopi/width
!    kvect=16*twopi/width
deltaf0=0._p_
  do i=1,total_number
     !deltaf0(i)=random_yj(0)
     deltaf0(i)=0.001*sin(kvect*r(i))*exp(-v(i)**2)/sqrt(pi)
     !deltaf0(i)=0.001*(sin(kvect*r(i))+cos(kvect*r(i)))*exp(-v(i)**2)/sqrt(pi)
     !deltaf0(i)=0.001*(cos(kvect*r(i)))*exp(-v(i)**2)/sqrt(pi)
     !if(v(i).le.0._p_) deltaf0(i)=2*0.001*(cos(kvect*r(i)))*exp(-v(i)**2)/sqrt(pi)
     !deltaf0(i)=0.0005*sin(kvect*r(i))*width
     !deltaf0(i)=0.00
  enddo
  write(*,*) 'kvect=',kvect

end subroutine set_deltaf0


subroutine  identify_rp(total_number,r,v,deltaf0,deltaf) !identify resonant particles
! select the top num markers, for which the variation/increase of deltaf is large
  use precision,only:p_
  use constants,only: twopi,pi
  implicit none
  integer,intent(in):: total_number
  real(p_),intent(in):: r(total_number),v(total_number),deltaf(total_number),deltaf0(total_number)
  integer,parameter:: num=500
  integer:: k(1),n,i
  real(p_)::deltaf_val(total_number),min_val
  real(p_):: r_top,v_top

  deltaf_val=abs(deltaf-deltaf0)
!  deltaf_val=deltaf-deltaf0
!  deltaf_val=deltaf0-deltaf
  do i=1,total_number
     deltaf_val(i)=deltaf_val(i)/(exp(-v(i)**2)/sqrt(pi))
    ! deltaf_val(i)=deltaf_val(i)
  enddo
  min_val=minval(deltaf_val)-1.

  open(11,file='top_particles.txt')
  do n=1,num
     k=maxloc(deltaf_val)
     r_top=r(k(1))
     v_top=v(k(1))
     write(11,'(i5,4(1pe14.5))') n, r_top,v_top,deltaf_val(k(1)),(deltaf(k(1))-deltaf0(k(1)))/(exp(-v(k(1))**2)/sqrt(pi))
     deltaf_val(k(1))=min_val !set this to a small value 
  enddo
  close(11)

end subroutine identify_rp
